import json

from mlfmu.types.fmu_component import FmiModel
from mlfmu.utils.builder import validate_interface_spec
from mlfmu.utils.fmi_builder import generate_model_description


def test_generate_simple_model_description():
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [{"name": "input1", "description": "My input1", "agentInputIndexes": ["0"], "type": "integer"}],
        "outputs": [{"name": "output1", "description": "My output1", "agentInputIndexes": ["0"]}],
    }
    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)
    xml_structure = generate_model_description(fmu_model=fmi_model)
    variables = xml_structure.findall(".//ScalarVariable")

    assert xml_structure.getroot().tag == "fmiModelDescription"
    assert variables[0].attrib["name"] == "input1"
    assert variables[0].attrib["causality"] == "input"
    assert variables[0].attrib["variability"] == "continuous"
    assert variables[0].attrib["description"] == "My input1"
    assert variables[0][0].tag == "Integer"

    assert variables[1].attrib["name"] == "output1"
    assert variables[1].attrib["causality"] == "output"
    assert variables[1].attrib["variability"] == "continuous"
    assert variables[1].attrib["description"] == "My output1"
    assert variables[1][0].tag == "Real"


def test_generate_model_description_with_internal_state_params():
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "states": [
            {
                "name": "state1",
                "description": "My state1",
                "startValue": 0.0,
                "type": "real",
                "agentOutputIndexes": ["0"],
            }
        ],
        "outputs": [{"name": "output1", "description": "My output1", "agentInputIndexes": ["0"]}],
    }
    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)
    xml_structure = generate_model_description(fmu_model=fmi_model)
    variables = xml_structure.findall(".//ScalarVariable")

    assert xml_structure.getroot().tag == "fmiModelDescription"
    assert variables[0].attrib["name"] == "state1"
    assert variables[0].attrib["causality"] == "parameter"
    assert variables[0][0].tag == "Real"
    assert variables[0][0].attrib["start"] == "0.0"

    assert variables[1].attrib["name"] == "output1"
    assert variables[1].attrib["causality"] == "output"


def test_generate_vector_ports():
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {
                "name": "inputVector",
                "description": "My input1",
                "agentInputIndexes": ["0:5"],
                "type": "real",
                "isArray": True,
                "length": 5,
            }
        ],
        "outputs": [
            {
                "name": "outputVector",
                "description": "My output1",
                "agentInputIndexes": ["0:5"],
                "isArray": True,
                "length": 5,
            }
        ],
    }
    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)
    xml_structure = generate_model_description(fmu_model=fmi_model)
    variables = xml_structure.findall(".//ScalarVariable")

    assert model
    assert variables[0].attrib["name"] == "inputVector[0]"
    assert variables[1].attrib["name"] == "inputVector[1]"
    assert variables[2].attrib["name"] == "inputVector[2]"
    assert variables[3].attrib["name"] == "inputVector[3]"
    assert variables[4].attrib["name"] == "inputVector[4]"

    assert variables[5].attrib["name"] == "outputVector[0]"
    assert variables[6].attrib["name"] == "outputVector[1]"
    assert variables[7].attrib["name"] == "outputVector[2]"
    assert variables[8].attrib["name"] == "outputVector[3]"
    assert variables[9].attrib["name"] == "outputVector[4]"


def test_generate_model_description_with_start_value():
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "usesTime": True,
        "inputs": [
            {
                "name": "input1",
                "description": "My input1",
                "agentInputIndexes": ["0"],
                "type": "integer",
                "startValue": 10,
            },
            {
                "name": "input2",
                "description": "My input2",
                "agentOutputIndexes": ["0"],
                "type": "boolean",
                "startValue": True,
            },
            {"name": "input3", "description": "My input3", "agentOutputIndexes": ["0"], "startValue": 10.0},
        ],
    }
    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)
    xml_structure = generate_model_description(fmu_model=fmi_model)
    variables = xml_structure.findall(".//ScalarVariable")

    assert xml_structure.getroot().tag == "fmiModelDescription"
    assert variables[0].attrib["name"] == "input1"
    assert variables[0].attrib["causality"] == "input"
    assert variables[0][0].tag == "Integer"
    assert variables[0][0].attrib["start"] == "10"

    assert variables[1].attrib["name"] == "input2"
    assert variables[1].attrib["causality"] == "input"
    assert variables[1][0].tag == "Boolean"
    assert variables[1][0].attrib["start"] == "True"

    assert variables[2].attrib["name"] == "input3"
    assert variables[2].attrib["causality"] == "input"
    assert variables[2][0].tag == "Real"
    assert variables[2][0].attrib["start"] == "10.0"


def test_generate_model_description_output():
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "usesTime": True,
        "inputs": [
            {
                "name": "input1",
                "description": "My input1",
                "agentInputIndexes": ["0"],
                "type": "integer",
                "startValue": 10,
            },
            {
                "name": "input2",
                "description": "My input2",
                "agentOutputIndexes": ["0"],
                "type": "boolean",
                "startValue": True,
            },
            {"name": "input3", "description": "My input3", "agentOutputIndexes": ["0"], "startValue": 10.0},
        ],
        "outputs": [
            {"name": "output1", "description": "My output1", "agentInputIndexes": ["0"], "type": "real"},
            {"name": "output1", "description": "My output1", "agentInputIndexes": ["0"], "type": "real"},
        ],
    }
    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)
    xml_structure = generate_model_description(fmu_model=fmi_model)
    variables = xml_structure.findall(".//ScalarVariable")
    output_variables = [var for var in variables if var.attrib.get("causality") == "output"]
    outputs_registered = xml_structure.findall(".//Outputs/Unknown")

    assert output_variables[0].attrib["valueReference"] == outputs_registered[0].attrib["index"]
    assert output_variables[1].attrib["valueReference"] == outputs_registered[1].attrib["index"]



import logging
import os
import shutil
import subprocess
from pathlib import Path

from pydantic import ValidationError

from mlfmu.types.fmu_component import FmiModel, ModelComponent
from mlfmu.types.onnx_model import ONNXModel
from mlfmu.utils.fmi_builder import generate_model_description
from mlfmu.utils.signals import range_list_expanded

# Paths to files needed for build
path_to_this_file = Path(__file__).resolve()
absolute_path = path_to_this_file.parent.parent
fmu_build_folder = absolute_path / "fmu_build"
template_parent_path = fmu_build_folder / "templates" / "fmu"

logger = logging.getLogger(__name__)


def format_template_file(
    template_path: Path,
    save_path: Path,
    data: dict[str, str],
) -> None:
    """
    Replace all the template strings with their corresponding values and save to a new file.

    Args
    ----
        template_path (Path): The path to the template file.
        save_path (Path): The path to save the formatted file.
        data (dict[str, str]): The data containing the values to replace in the template.
    """
    # TODO: Need to check that these calls are safe from a cybersecurity point of view  # noqa: TD002
    with Path.open(template_path, encoding="utf-8") as template_file:
        template_string = template_file.read()

    formatted_string = template_string.format(**data)
    with Path.open(save_path, "w", encoding="utf-8") as save_file:
        _ = save_file.write(formatted_string)


def create_model_description(
    fmu: FmiModel,
    src_path: Path,
) -> None:
    """
    Generate modelDescription.xml structure for FMU, and save it in a file.

    Args
    ----
        fmu (FmiModel): The FMI model.
        src_path (Path): The path to save the model description file.
    """
    xml_structure = generate_model_description(fmu_model=fmu)

    # Save in file
    xml_structure.write(src_path / "modelDescription.xml", encoding="utf-8")


def make_fmu_dirs(src_path: Path) -> None:
    """
    Create all the directories needed to put all the FMU files in.

    Args
    ----
        src_path (Path): The path to the FMU source directory.
    """
    sources_path = src_path / "sources"
    resources_path = src_path / "resources"
    sources_path.mkdir(parents=True, exist_ok=True)
    resources_path.mkdir(parents=True, exist_ok=True)


def create_files_from_templates(
    data: dict[str, str],
    fmu_src: Path,
) -> None:
    """
    Create and format all needed C++ files for FMU generation.

    Args
    ----
        data (dict[str, str]): The data containing the values to format the template files.
        fmu_src (Path): The path to the FMU source directory.
    """
    sources_path = fmu_src / "sources"
    file_names = ["fmu.cpp", "model_definitions.h"]

    paths = [
        (
            template_parent_path / "_template.".join(file_name.split(".")),
            sources_path / file_name,
        )
        for file_name in file_names
    ]

    for template_path, save_path in paths:
        format_template_file(template_path, save_path, data)


def format_template_data(onnx: ONNXModel, fmi_model: FmiModel, model_component: ModelComponent) -> dict[str, str]:
    """
    Generate the key-value pairs needed to format the template files to valid C++.

    Args
    ----
        onnx (ONNXModel): The ONNX model.
        fmi_model (FmiModel): The FMI model.
        model_component (ModelComponent): The model component.

    Returns
    -------
        dict[str, str]: The formatted template data.
    """
    # Work out template mapping between ONNX and FMU ports
    inputs, outputs, state_init = fmi_model.get_template_mapping()
    state_output_indexes = [
        index for state in model_component.states for index in range_list_expanded(state.agent_output_indexes)
    ]

    # Total number of inputs/outputs/internal states
    num_fmu_variables = fmi_model.get_total_variable_number()
    num_fmu_inputs = len(inputs)
    num_fmu_outputs = len(outputs)
    num_onnx_states = len(state_output_indexes)
    num_onnx_state_init = len(state_init)

    # Checking compatibility between ModelComponent and ONNXModel
    if num_fmu_inputs > onnx.input_size:
        raise ValueError(
            "The number of total input indexes for all inputs and parameter in the interface file "
            f"(={num_fmu_inputs}) cannot exceed the input size of the ml model (={onnx.input_size})"
        )

    if num_fmu_outputs > onnx.output_size:
        raise ValueError(
            "The number of total output indexes for all outputs in the interface file "
            f"(={num_fmu_outputs}) cannot exceed the output size of the ml model (={onnx.output_size})"
        )
    if num_onnx_states > onnx.state_size:
        raise ValueError(
            "The number of total output indexes for all states in the interface file "
            f"(={num_onnx_states}) cannot exceed either the state input size (={onnx.state_size})"
        )
    if num_onnx_state_init > onnx.state_size:
        raise ValueError(
            "The number of states that are initialized in the interface file "
            f"(={num_onnx_state_init}) cannot exceed either the state input size (={onnx.state_size})"
        )

    # Flatten vectors to comply with template requirements
    # -> onnx-index, variable-reference, onnx-index, variable-reference ...
    flattened_input_string = ", ".join(
        [str(index) for indexValueReferencePair in inputs for index in indexValueReferencePair]
    )
    flattened_output_string = ", ".join(
        [str(index) for indexValueReferencePair in outputs for index in indexValueReferencePair]
    )
    flattened_state_string = ", ".join([str(index) for index in state_output_indexes])
    flattened_state_init_string = ", ".join(
        [str(index) for indexValueReferencePair in state_init for index in indexValueReferencePair]
    )

    template_data: dict[str, str] = {
        "numFmuVariables": str(num_fmu_variables),
        "FmuName": fmi_model.name,
        "numOnnxInputs": str(onnx.input_size),
        "numOnnxOutputs": str(onnx.output_size),
        "numOnnxStates": str(onnx.state_size),
        "numOnnxStateInit": str(num_onnx_state_init),
        "onnxUsesTime": "true" if onnx.time_input else "false",
        "onnxInputName": onnx.input_name,
        "onnxStatesName": onnx.states_name,
        "onnxTimeInputName": onnx.time_input_name,
        "onnxOutputName": onnx.output_name,
        "onnxFileName": onnx.filename,
        "numOnnxFmuInputs": str(num_fmu_inputs),
        "numOnnxFmuOutputs": str(num_fmu_outputs),
        "numOnnxStatesOutputs": str(num_onnx_states),
        "onnxInputValueReferences": flattened_input_string,
        "onnxOutputValueReferences": flattened_output_string,
        "onnxStateOutputIndexes": flattened_state_string,
        "onnxStateInitValueReferences": flattened_state_init_string,
    }

    return template_data


def validate_interface_spec(
    spec: str,
) -> tuple[ValidationError | None, ModelComponent]:
    """
    Parse and validate JSON data from the interface file.

    Args
    ----
        spec (str): Contents of the JSON file.

    Returns
    -------
        tuple[Optional[ValidationError], ModelComponent]:
            The validation error (if any) and the validated model component.
            The pydantic model instance that contains all the interface information.
    """
    parsed_spec = ModelComponent.model_validate_json(json_data=spec, strict=True)
    try:
        validated_model = ModelComponent.model_validate(parsed_spec)
    except ValidationError as e:
        return e, parsed_spec

    return None, validated_model


def generate_fmu_files(
    fmu_src_path: os.PathLike[str],
    onnx_path: os.PathLike[str],
    interface_spec_path: os.PathLike[str],
) -> FmiModel:
    """
    Generate FMU files based on the FMU source, ONNX model, and interface specification.

    Args
    ----
        fmu_src_path (os.PathLike[str]): The path to the FMU source directory.
        onnx_path (os.PathLike[str]): The path to the ONNX model file.
        interface_spec_path (os.PathLike[str]): The path to the interface specification file.

    Returns
    -------
        FmiModel: The FMI model.
    """
    # Create Path instances for the path to the spec and ONNX file.
    onnx_path = Path(onnx_path)
    interface_spec_path = Path(interface_spec_path)

    # Load JSON interface contents
    with Path.open(interface_spec_path, encoding="utf-8") as template_file:
        interface_contents = template_file.read()

    # Validate the FMU interface spec against expected Schema
    error, component_model = validate_interface_spec(interface_contents)

    if error:
        # Display error and finish workflow
        raise error

    # Create ONNXModel and FmiModel instances -> load some metadata
    onnx_model = ONNXModel(onnx_path=onnx_path, time_input=bool(component_model.uses_time))
    fmi_model = FmiModel(model=component_model)
    fmu_source = Path(fmu_src_path) / fmi_model.name

    template_data = format_template_data(onnx=onnx_model, fmi_model=fmi_model, model_component=component_model)

    # Generate all FMU files
    make_fmu_dirs(fmu_source)
    create_files_from_templates(data=template_data, fmu_src=fmu_source)
    create_model_description(fmu=fmi_model, src_path=fmu_source)

    # Copy ONNX file and save it inside FMU folder
    _ = shutil.copyfile(src=onnx_path, dst=fmu_source / "resources" / onnx_model.filename)

    return fmi_model


def validate_fmu_source_files(fmu_path: os.PathLike[str]) -> None:
    """
    Validate the FMU source files.

    Args
    ----
        fmu_path (os.PathLike[str]): The path to the FMU source directory.

    Raises
    ------
        FileNotFoundError: If required files are missing in the FMU source directory.
    """
    fmu_path = Path(fmu_path)

    files_should_exist: list[str] = [
        "modelDescription.xml",
        "sources/fmu.cpp",
        "sources/model_definitions.h",
    ]

    if files_not_exists := [file for file in files_should_exist if not (fmu_path / file).is_file()]:
        raise FileNotFoundError(
            f"The files {files_not_exists} are not contained in the provided FMU source path ({fmu_path})"
        )

    resources_dir = fmu_path / "resources"

    num_onnx_files = len(list(resources_dir.glob("*.onnx")))

    if num_onnx_files < 1:
        raise FileNotFoundError(
            f"There is no *.onnx file in the resource folder in the provided FMU source path ({fmu_path})"
        )


def build_fmu(
    fmu_src_path: os.PathLike[str],
    fmu_build_path: os.PathLike[str],
    fmu_save_path: os.PathLike[str],
) -> None:
    """
    Build the FMU.

    Args
    ----
        fmu_src_path (os.PathLike[str]): The path to the FMU source directory.
        fmu_build_path (os.PathLike[str]): The path to the FMU build directory.
        fmu_save_path (os.PathLike[str]): The path to save the built FMU.

    Raises
    ------
        FileNotFoundError: If required files are missing in the FMU source directory.
    """
    fmu_src_path = Path(fmu_src_path)
    validate_fmu_source_files(fmu_src_path)
    fmu_name = fmu_src_path.stem
    conan_install_command = [
        "conan",
        "install",
        ".",
        "-of",
        str(fmu_build_path),
        "-u",
        "-b",
        "missing",
        "-o",
        "shared=True",
    ]
    cmake_set_folders = [
        f"-DCMAKE_BINARY_DIR={fmu_build_path!s}",
        f"-DFMU_OUTPUT_DIR={fmu_save_path!s}",
        f"-DFMU_NAMES={fmu_name}",
        f"-DFMU_SOURCE_PATH={fmu_src_path.parent!s}",
    ]
    cmake_command = ["cmake", *cmake_set_folders, "--preset", "conan-default"]
    cmake_build_command = ["cmake", "--build", ".", "-j", "14", "--config", "Release"]

    # Change directory to the build folder
    os.chdir(fmu_build_folder)

    # Run conan install, cmake, cmake build
    logger.debug("Builder: Run conan install")
    try:
        _ = subprocess.run(conan_install_command, check=True)  # noqa: S603
    except subprocess.CalledProcessError:
        logger.exception("Exception in conan install: %s")

    logger.debug("Builder: Run cmake")
    try:
        _ = subprocess.run(cmake_command, check=True)  # noqa: S603
    except subprocess.CalledProcessError:
        logger.exception("Exception in cmake: %s")

    os.chdir(fmu_build_path)
    logger.debug("Builder: Run cmake build")
    try:
        _ = subprocess.run(cmake_build_command, check=True)  # noqa: S603
    except subprocess.CalledProcessError:
        logger.exception("Exception in cmake build: %s")

    logger.debug("Builder: Done with build_fmu")

    # Return to original working directory (leave build dir)
    os.chdir(absolute_path)



# pyright: reportPrivateUsage=false
import sys
from argparse import ArgumentError
from dataclasses import dataclass, field
from pathlib import Path

import pytest

from mlfmu.api import MlFmuCommand
from mlfmu.cli import mlfmu
from mlfmu.cli.mlfmu import _argparser, main

# *****Test commandline interface (CLI)************************************************************


@dataclass()
class CliArgs:
    # Expected default values for the CLI arguments when mlfmu gets called via the commandline
    quiet: bool = False
    verbose: bool = False
    log: str | None = None
    log_level: str = field(default_factory=lambda: "WARNING")
    command: str = ""


@pytest.mark.parametrize(
    "inputs, expected",
    [
        ([], ArgumentError),
        (["asd"], ArgumentError),
        (["build", "-q"], CliArgs(quiet=True, command="build")),
        (["build", "--quiet"], CliArgs(quiet=True, command="build")),
        (["build", "-v"], CliArgs(verbose=True, command="build")),
        (["build", "--verbose"], CliArgs(verbose=True, command="build")),
        (["build", "-qv"], ArgumentError),
        (["build", "--log", "logFile"], CliArgs(log="logFile", command="build")),
        (["build", "--log"], ArgumentError),
        (["build", "--log-level", "INFO"], CliArgs(log_level="INFO", command="build")),
        (["build", "--log-level"], ArgumentError),
        (["build", "-o"], ArgumentError),
    ],
)
def test_cli(
    inputs: list[str],
    expected: CliArgs | type,
    monkeypatch: pytest.MonkeyPatch,
):
    """
    Test the command-line interface (CLI) of the 'mlfmu' program.

    Args
    ----
        inputs (List[str]): A list of input arguments to be passed to the CLI.
        expected (Union[CliArgs, type]): The expected output of the CLI.
            It can be either an instance of the `CliArgs` class or a subclass of `Exception`.
        monkeypatch (pytest.MonkeyPatch): A pytest fixture that allows patching of objects at runtime.

    Raises
    ------
        AssertionError: If the `expected` argument is neither an instance of `CliArgs`
            nor a subclass of `Exception`.
    """

    # sourcery skip: no-conditionals-in-tests
    # sourcery skip: no-loop-in-tests

    # Prepare
    monkeypatch.setattr(sys, "argv", ["mlfmu", *inputs])
    parser = _argparser()

    # Execute
    if isinstance(expected, CliArgs):
        args_expected: CliArgs = expected
        args = parser.parse_args()

        # Assert args
        for key in args_expected.__dataclass_fields__:
            assert args.__getattribute__(key) == args_expected.__getattribute__(key)
    elif issubclass(expected, Exception):
        exception: type = expected

        # Assert that expected exception is raised
        with pytest.raises((exception, SystemExit)):
            args = parser.parse_args()
    else:
        raise AssertionError


# *****Ensure the CLI correctly configures logging*************************************************


@dataclass()
class ConfigureLoggingArgs:
    # Values that main() is expected to pass to ConfigureLogging() by default when configuring the logging
    log_level_console: str = field(default_factory=lambda: "WARNING")
    log_file: Path | None = None
    log_level_file: str = field(default_factory=lambda: "WARNING")


@pytest.mark.parametrize(
    "inputs, expected",
    [
        ([], ArgumentError),
        (["build"], ConfigureLoggingArgs()),
        (["build", "-q"], ConfigureLoggingArgs(log_level_console="ERROR")),
        (
            ["build", "--quiet"],
            ConfigureLoggingArgs(log_level_console="ERROR"),
        ),
        (["build", "-v"], ConfigureLoggingArgs(log_level_console="INFO")),
        (
            ["build", "--verbose"],
            ConfigureLoggingArgs(log_level_console="INFO"),
        ),
        (["build", "-qv"], ArgumentError),
        (
            ["build", "--log", "logFile"],
            ConfigureLoggingArgs(log_file=Path("logFile")),
        ),
        (["build", "--log"], ArgumentError),
        (
            ["build", "--log-level", "INFO"],
            ConfigureLoggingArgs(log_level_file="INFO"),
        ),
        (["build", "--log-level"], ArgumentError),
    ],
)
def test_logging_configuration(
    inputs: list[str],
    expected: ConfigureLoggingArgs | type,
    monkeypatch: pytest.MonkeyPatch,
):
    """
    Test the logging configuration of the `main` function in the `mlfmu` module.

    Args
    ----
        inputs (List[str]): The list of input arguments to be passed to the `main` function.
        expected (Union[ConfigureLoggingArgs, type]): The expected output of the `main` function.
        It can be an instance of `ConfigureLoggingArgs` or a subclass of `Exception`.
        monkeypatch (pytest.MonkeyPatch): The monkeypatch fixture provided by pytest.

    Raises
    ----
        AssertionError: If the `expected` argument is neither an instance of `ConfigureLoggingArgs`
        nor a subclass of `Exception`.
    """

    # sourcery skip: no-conditionals-in-tests
    # sourcery skip: no-loop-in-tests

    # Prepare
    monkeypatch.setattr(sys, "argv", ["mlfmu", *inputs])
    args: ConfigureLoggingArgs = ConfigureLoggingArgs()

    def fake_configure_logging(
        log_level_console: str,
        log_file: Path | None,
        log_level_file: str,
    ):
        args.log_level_console = log_level_console
        args.log_file = log_file
        args.log_level_file = log_level_file

    def fake_run(
        command: str,
        interface_file: str | None,
        model_file: str | None,
        fmu_path: str | None,
        source_folder: str | None,
    ):
        pass

    monkeypatch.setattr(mlfmu, "configure_logging", fake_configure_logging)
    monkeypatch.setattr(mlfmu, "run", fake_run)
    # Execute
    if isinstance(expected, ConfigureLoggingArgs):
        args_expected: ConfigureLoggingArgs = expected
        main()
        # Assert args
        for key in args_expected.__dataclass_fields__:
            assert args.__getattribute__(key) == args_expected.__getattribute__(key)
    elif issubclass(expected, Exception):
        exception: type = expected
        # Assert that expected exception is raised
        with pytest.raises((exception, SystemExit)):
            main()
    else:
        raise AssertionError


# *****Ensure the CLI correctly invokes the API****************************************************


@dataclass()
class ApiArgs:
    # Values that main() is expected to pass to run() by default when invoking the API
    command: MlFmuCommand | None = None
    interface_file: str | None = None
    model_file: str | None = None
    fmu_path: str | None = None
    source_folder: str | None = None


@pytest.mark.parametrize(
    "inputs, expected",
    [
        ([], ArgumentError),
        (["build"], ApiArgs()),
    ],
)
def test_api_invokation(
    inputs: list[str],
    expected: ApiArgs | type,
    monkeypatch: pytest.MonkeyPatch,
):
    # sourcery skip: no-conditionals-in-tests
    # sourcery skip: no-loop-in-tests
    """
    Test the invocation of the API function.

    Args
    ----
        inputs (List[str]): The list of input arguments.
        expected (Union[ApiArgs, type]): The expected output, either an instance of ApiArgs or an exception type.
        monkeypatch (pytest.MonkeyPatch): The monkeypatch object for patching sys.argv.

    Raises
    ----
        AssertionError: If the expected output is neither an instance of ApiArgs nor an exception type.
    """

    # Prepare
    monkeypatch.setattr(sys, "argv", ["mlfmu", *inputs])
    args: ApiArgs = ApiArgs()

    def fake_run(
        command: str,
        interface_file: str | None,
        model_file: str | None,
        fmu_path: str | None,
        source_folder: str | None,
    ):
        args.command = MlFmuCommand.from_string(command)
        args.interface_file = interface_file
        args.model_file = model_file
        args.fmu_path = fmu_path
        args.source_folder = source_folder

    monkeypatch.setattr(mlfmu, "run", fake_run)
    # Execute
    if isinstance(expected, ApiArgs):
        args_expected: ApiArgs = expected
        main()
        # Assert args
        for key in args_expected.__dataclass_fields__:
            assert args.__getattribute__(key) == args_expected.__getattribute__(key)
    elif issubclass(expected, Exception):
        exception: type = expected
        # Assert that expected exception is raised
        with pytest.raises((exception, SystemExit)):
            main()
    else:
        raise AssertionError



import os
from pathlib import Path
from typing import Any

from onnxruntime import InferenceSession, NodeArg


class ONNXModel:
    """ONNX Metadata class.

    Represents an ONNX model and provides methods to load inputs and outputs.
    Allows to import the ONNX file and figure out the input/output sizes.

    Raises
    ------
    ValueError
        If the ml model has 3 inputs, but the `usesTime` flag is set to `false` in the json interface.
    ValueError
        If the number of inputs to the ml model is larger than 3.
    ValueError
        If the number of outputs from the ml model is not exactly 1.
    """

    filename: str = ""  #: The name of the ONNX file.
    states_name: str = ""  #: The name of the internal states input.
    state_size: int = 0  #: The size of the internal states input.
    input_name: str = ""  #: The name of the main input.
    input_size: int = 0  #: The size of the main input.
    output_name: str = ""  #: The name of the output.
    output_size: int = 0  #: The size of the output.
    time_input_name: str = ""  #: The name of the time input.
    time_input: bool = False  #: Flag indicating whether the model uses time input.
    __onnx_path: Path  #: The path to the ONNX file.
    __onnx_session: InferenceSession  #: The ONNX runtime inference session.

    def __init__(
        self,
        onnx_path: str | os.PathLike[str],
        *,
        time_input: bool = False,
    ) -> None:
        """
        Initialize the ONNXModel object by loading the ONNX file and assigning model parameters.

        Args:
            onnx_path (Union[str, os.PathLike[str]]): The path to the ONNX file.
            time_input (bool, optional): Flag indicating whether the model uses time input. Defaults to False.
        """
        # Load ONNX file into memory
        self.__onnx_path = onnx_path if isinstance(onnx_path, Path) else Path(onnx_path)
        self.__onnx_session = InferenceSession(onnx_path)

        # Assign model parameters
        self.filename = f"{self.__onnx_path.stem}.onnx"
        self.time_input = time_input

        self.load_inputs()
        self.load_outputs()

    def load_inputs(self) -> None:
        """Load the inputs from the ONNX file and assign the input name and size."""
        # Get inputs from ONNX file
        inputs: list[NodeArg] = self.__onnx_session.get_inputs()
        input_names = [inp.name for inp in inputs]
        input_shapes = [inp.shape for inp in inputs]
        self.input_name = input_names[0]
        self.input_size = input_shapes[0][1]

        # Number of internal states
        num_states = 0

        # Based on number of inputs work out which are INTERNAL STATES, INPUTS and TIME DATA
        if len(input_names) == 3:  # noqa: PLR2004
            self.states_name = input_names[1]
            self.time_input_name = input_names[2]
            num_states = input_shapes[1][1]
            if not self.time_input:
                raise ValueError(
                    "The ml model has 3 inputs, but the `usesTime` flag is set to `false` in the json interface. "
                    "A model can only have 3 inputs if it uses time input."
                )
        elif len(input_names) == 2:  # noqa: PLR2004
            if self.time_input:
                self.time_input_name = input_names[1]
            else:
                self.states_name = input_names[1]
                num_states = input_shapes[1][1]

        elif not input_names or len(input_names) > 3:  # noqa: PLR2004
            raise ValueError(f"The number of inputs to the ml model (={len(input_names)}) must be 1, 2 or 3")

        self.state_size = num_states

    def load_outputs(self) -> None:
        """Load the outputs from the ONNX file and assign the output name and size."""
        # Get outputs from ONNX file
        outputs: list[Any] = self.__onnx_session.get_outputs()
        output_names = [out.name for out in outputs]

        if len(output_names) != 1:
            raise ValueError(f"The number of outputs from the ml model (={len(output_names)}) must be exactly 1")

        output_shapes = [out.shape for out in outputs]
        self.output_name = output_names[0]
        self.output_size = output_shapes[0][1]



from os import path
from conan import ConanFile
from conan.tools.build import can_run
from conan.tools.cmake import CMake, cmake_layout
from conan.tools.files import copy


class CppFmuTestConan(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
    generators = "CMakeDeps", "CMakeToolchain"
    test_type = "explicit"

    def requirements(self):
        self.requires(self.tested_reference_str)

    def layout(self):
        cmake_layout(self)

    def generate(self):
        for require, dep in self.dependencies.items():
            if require.build or require.test:
                continue
            if dep.ref.name == "cppfmu" and len(dep.cpp_info.srcdirs) > 0:
                copy(self, "fmi_functions.cpp", dep.cpp_info.srcdirs[0],
                    path.join(self.build_folder, dep.ref.name), keep_path=False)


    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def test(self):
        if can_run(self):
            bin_path = path.join(self.cpp.build.bindir, "tester")
            self.run(bin_path, env="conanrun")



import json
import os
from pathlib import Path
from typing import cast

from dictIO.utils.path import relative_path
from json_schema_for_humans.generate import (
    SchemaToRender,
    TemplateRenderer,
    generate_schemas_doc,
)
from json_schema_for_humans.generation_configuration import GenerationConfiguration
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass

from mlfmu.types.fmu_component import ModelComponent

__ALL__ = ["publish_interface_schema"]


def generate_interface_schema(
    model: BaseModel | ModelMetaclass,
    schema_dir: str | os.PathLike[str] | None = None,
) -> None:
    """
    Generate a JSON interface schema file for the given model.

    Args
    ----
        model (BaseModel | ModelMetaclass): The pydantic model for which to generate the schema.
        schema_dir (str | os.PathLike[str], optional):
            The directory where the schema file will be saved. Defaults to None.
    """
    schema_dir_default = Path.cwd() / "docs/schema"
    schema_dir = schema_dir or schema_dir_default
    # Make sure schema_dir argument is of type Path. If not, cast it to Path type.
    schema_dir = schema_dir if isinstance(schema_dir, Path) else Path(schema_dir)

    # Assert model argument is a pydantic BaseModel
    # Background: ModelMetaClass is added just to please static type checking,
    #             which would otherwise complain.
    #             Behind the scenes in pdyantic, models always inherit the attributes of BaseModel.
    if not hasattr(model, "model_json_schema"):
        raise ValueError("model argument must be a pydantic BaseModel")
    model = cast(BaseModel, model)

    # Create schema_dir if it does not exist
    schema_dir.mkdir(parents=True, exist_ok=True)

    json_file: Path = schema_dir / "schema.json"
    schema = json.dumps(model.model_json_schema(by_alias=True), indent=4)

    with Path.open(json_file, "w", encoding="utf-8") as f:
        _ = f.write(schema)

    return


def generate_interface_docs(
    schema_dir: str | os.PathLike[str] | None = None,
    docs_dir: str | os.PathLike[str] | None = None,
) -> None:
    """
    Generate HTML documentation for the JSON interface schema files in the schema directory.

    Args
    ----
        schema_dir (str | os.PathLike[str], optional):
            The directory where the schema files are located. Defaults to None.
        docs_dir (str | os.PathLike[str], optional):
            The directory where the documentation files will be saved. Defaults to None.
    """
    schema_dir_default = Path.cwd() / "docs/schema"
    schema_dir = schema_dir or schema_dir_default

    docs_dir_default = Path.cwd() / "docs/interface"
    docs_dir = docs_dir or docs_dir_default

    # Make sure schema_dir and docs_dir are of type Path. If not, cast it to Path type.
    schema_dir = schema_dir if isinstance(schema_dir, Path) else Path(schema_dir)
    docs_dir = docs_dir if isinstance(docs_dir, Path) else Path(docs_dir)

    # Create dirs in case they don't exist
    schema_dir.mkdir(parents=True, exist_ok=True)
    docs_dir.mkdir(parents=True, exist_ok=True)

    # Collect all schemata in schema dir
    pattern: str = "**.json"
    schemata: list[Path] = list(schema_dir.glob(pattern))

    # Generate html documentation for schemata
    config: GenerationConfiguration = GenerationConfiguration(
        template_name="js",
        expand_buttons=True,
        link_to_reused_ref=False,
        show_breadcrumbs=False,
    )

    schemas_to_render: list[SchemaToRender] = []

    for schema in schemata:
        rel_path: Path = relative_path(from_path=schema_dir, to_path=schema.parent)
        name: str = schema.stem
        html_file: Path = docs_dir / rel_path / f"{name}.html"
        schema_to_render: SchemaToRender = SchemaToRender(
            schema_file=schema,
            result_file=html_file,
            output_dir=None,
        )
        schemas_to_render.append(schema_to_render)

    _ = generate_schemas_doc(
        schemas_to_render=schemas_to_render,
        template_renderer=TemplateRenderer(config),
    )


def publish_interface_schema(
    schema_dir: str | os.PathLike[str] | None = None,
    docs_dir: str | os.PathLike[str] | None = None,
) -> None:
    """
    Publish the JSON schema and HTML documentation for the interface.

    Args
    ----
        schema_dir (str | os.PathLike[str], optional):
            The directory where the schema file will be saved. Defaults to None.
        docs_dir (str | os.PathLike[str], optional):
            The directory where the documentation files will be saved. Defaults to None.
    """
    # Generate JSON schema
    generate_interface_schema(model=ModelComponent, schema_dir=schema_dir)

    # Generate documentation HTML
    generate_interface_docs(schema_dir=schema_dir, docs_dir=docs_dir)



import json

import pytest
from pydantic import ValidationError

from mlfmu.types.fmu_component import ModelComponent
from mlfmu.utils.builder import validate_interface_spec


def test_validate_simple_interface_spec():
    # Assuming validate_interface_spec takes a dictionary as input
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [{"name": "input1", "description": "My input1", "agentInputIndexes": ["0"], "type": "integer"}],
        "outputs": [{"name": "output1", "description": "My output1", "agentInputIndexes": ["0"]}],
    }
    error, model = validate_interface_spec(json.dumps(valid_spec))
    assert error is None
    assert isinstance(model, ModelComponent)
    assert model.name == "example"
    assert model.version == "1.0"
    assert model.inputs[0].name == "input1"
    assert model.inputs[0].type == "integer"
    assert model.outputs[0].name == "output1"
    assert model.outputs[0].type == "real"


def test_validate_interface_spec_wrong_types():
    # Assuming validate_interface_spec returns False for invalid specs
    invalid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [{"name": "input1", "type": "enum"}],  # Missing enum type
        "outputs": [{"name": "output1", "type": "int"}],  # Should be integer
    }

    with pytest.raises(ValidationError) as exc_info:
        _, _ = validate_interface_spec(json.dumps(invalid_spec))

    # Model type error as it's missing the agentInputIndexes
    assert exc_info.match("Input should be 'real', 'integer', 'string' or 'boolean'")


def test_validate_unnamed_spec():
    invalid_spec = {
        "version": "1.0",
        "inputs": [{"name": "input1", "description": "My input1", "agentInputIndexes": ["0"], "type": "integer"}],
        "outputs": [{"name": "output1", "description": "My output1", "agentInputIndexes": ["0"]}],
    }

    with pytest.raises(ValidationError) as exc_info:
        _, _ = validate_interface_spec(json.dumps(invalid_spec))

    assert exc_info.match("Field required")


def test_validate_invalid_agent_indices():
    invalid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "input1", "description": "My input1", "type": "integer", "agentInputIndexes": [0, ":10", "10.0"]}
        ],  # Should be a stringified list of integers
        "outputs": [
            {"name": "output1", "description": "My output1", "agentOutputIndexes": ["0:a"]}
        ],  # Should not have letters
    }

    with pytest.raises(ValidationError) as exc_info:
        _, _ = validate_interface_spec(json.dumps(invalid_spec))

    assert exc_info.match("Input should be a valid string")
    assert exc_info.match("String should match pattern")
    assert exc_info.match("4 validation errors for ModelComponent")


def test_validate_default_parameters():
    invalid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "input1", "description": "My input1", "type": "integer", "agentInputIndexes": ["10"]}
        ],  # Should be a stringified list of integers
        "outputs": [
            {"name": "output1", "description": "My output1", "agentOutputIndexes": ["0:10"]}
        ],  # Should not have letters
    }
    error, model = validate_interface_spec(json.dumps(invalid_spec))
    assert error is None
    assert model is not None

    assert model.uses_time is False
    assert model.state_initialization_reuse is False
    assert model.name == "example"
    assert model.parameters == []


def test_validate_internal_states():
    invalid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "input1", "description": "My input1", "type": "integer", "agentInputIndexes": ["10"]}
        ],  # Should be a stringified list of integers
        "outputs": [
            {"name": "output1", "description": "My output1", "agentOutputIndexes": ["0:10"]}
        ],  # Should not have letters
        "states": [
            {
                "name": "state1",
                "description": "My state1",
                "startValue": 10,
                "initializationVariable": "input1",
                "agentOutputIndexes": ["0:10"],
            },
            {"name": "state2", "description": "My state2", "agentOutputIndexes": ["0:10"]},
            {"name": "state3", "initializationVariable": "input1", "agentOutputIndexes": ["0:10"]},
            {"description": "My state4", "startValue": 10},
        ],
    }
    with pytest.raises(ValidationError) as exc_info:
        _, _ = validate_interface_spec(json.dumps(invalid_spec))

    assert exc_info.match(
        "Value error, Only one state initialization method is allowed to be used at a time: \
initialization_variable cannot be set if either start_value or name is set."
    )
    assert exc_info.match(
        "Value error, name is set without start_value being set. \
Both fields need to be set for the state initialization to be valid"
    )
    assert exc_info.match(
        "Value error, Only one state initialization method is allowed to be used at a time: \
initialization_variable cannot be set if either start_value or name is set."
    )
    assert exc_info.match(
        "Value error, start_value is set without name being set. \
Both fields need to be set for the state initialization to be valid"
    )



import os
from pathlib import Path


def find_default_file(
    folder: Path,
    file_extension: str,
    default_name: str | None = None,
) -> Path | None:
    """Return a file inside folder with the file extension that matches file_extension.

    If there are multiple matches it uses the closest match to default_name if given.
    Return None if there is no clear match.

    Parameters
    ----------
    folder : Path
        the folder to search in
    file_extension : str
        file extension to search for
    default_name : str | None, optional
        file name used to determine "closest match"
        in case multiple files match file_extension, by default None

    Returns
    -------
    Path | None
        the path to the file if it is found, otherwise None
    """
    # Check if there is a file with correct file extension in current working directory. If it exists use it.
    matching_files: list[Path] = []

    for file in os.listdir(folder):
        file_path = folder / file
        if file_path.is_file() and file_path.suffix.lstrip(".") == file_extension:
            matching_files.append(file_path)

    if not matching_files:
        return None

    if len(matching_files) == 1:
        return matching_files[0]

    # If there are more matches on file extension. Use the one that matches the default name
    if default_name is None:
        return None

    name_matches = [file for file in matching_files if default_name in file.stem]

    if not name_matches:
        return None

    if len(name_matches) == 1:
        return name_matches[0]

    # If more multiple name matches use the exact match if it exists
    name_exact_matches = [file for file in matching_files if default_name == file.stem]

    return name_matches[0] if len(name_exact_matches) == 1 else None



import json
import re
from pathlib import Path

import pytest

from mlfmu.types.fmu_component import FmiModel
from mlfmu.types.onnx_model import ONNXModel
from mlfmu.utils.builder import format_template_data, validate_interface_spec


@pytest.fixture(scope="session")
def wind_generator_onnx() -> ONNXModel:
    return ONNXModel(Path.cwd().parent / "data" / "example.onnx", time_input=True)


def test_valid_template_data(wind_generator_onnx: ONNXModel):
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "inputs", "description": "My inputs", "agentInputIndexes": ["0:2"], "isArray": True, "length": 2}
        ],
        "outputs": [
            {
                "name": "outputs",
                "description": "My outputs",
                "agentOutputIndexes": ["0:2"],
                "isArray": True,
                "length": 2,
            }
        ],
        "states": [
            {"agentOutputIndexes": ["2:130"]},
            {"name": "state1", "startValue": 10.0, "agentOutputIndexes": ["0"]},
            {"name": "state2", "startValue": 180.0, "agentOutputIndexes": ["1"]},
        ],
    }
    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)
    template_data = format_template_data(onnx=wind_generator_onnx, fmi_model=fmi_model, model_component=model)

    assert template_data["FmuName"] == "example"
    assert template_data["numFmuVariables"] == "6"
    assert template_data["numOnnxInputs"] == "2"
    assert template_data["numOnnxOutputs"] == "130"
    assert template_data["numOnnxStates"] == "130"
    assert template_data["onnxInputValueReferences"] == "0, 0, 1, 1"
    assert template_data["onnxOutputValueReferences"] == "0, 2, 1, 3"


def test_template_data_invalid_input_size(wind_generator_onnx: ONNXModel):
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "inputs", "description": "My inputs", "agentInputIndexes": ["0:2"], "isArray": True, "length": 2},
            {
                "name": "inputs2",
                "description": "My inputs 2",
                "agentInputIndexes": ["0:10"],
                "isArray": True,
                "length": 10,
            },
        ],
        "outputs": [
            {"name": "outputs", "description": "My outputs", "agentInputIndexes": ["0:2"], "isArray": True, "length": 2}
        ],
        "states": [
            {"agentOutputIndexes": ["2:130"]},
            {"name": "state1", "startValue": 10.0, "agentOutputIndexes": ["0"]},
            {"name": "state2", "startValue": 180.0, "agentOutputIndexes": ["1"]},
        ],
    }

    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)

    with pytest.raises(ValueError) as exc_info:
        _ = format_template_data(onnx=wind_generator_onnx, fmi_model=fmi_model, model_component=model)

    assert exc_info.match(
        re.escape(
            "The number of total input indexes for all inputs and parameter in the interface file (=12) \
cannot exceed the input size of the ml model (=2)"
        )
    )


def test_template_data_invalid_output_size(wind_generator_onnx: ONNXModel):
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "inputs", "description": "My inputs", "agentInputIndexes": ["0:2"], "isArray": True, "length": 2}
        ],
        "outputs": [
            {
                "name": "outputs",
                "description": "My outputs",
                "agentOutputIndexes": ["0:2"],
                "isArray": True,
                "length": 2,
            },
            {
                "name": "outputs2",
                "description": "My outputs 2",
                "agentOutputIndexes": ["0:200"],
                "isArray": True,
                "length": 200,
            },
        ],
        "states": [
            {"agentOutputIndexes": ["2:130"]},
            {"name": "state1", "startValue": 10.0, "agentOutputIndexes": ["0"]},
            {"name": "state2", "startValue": 180.0, "agentOutputIndexes": ["1"]},
        ],
    }

    _, model = validate_interface_spec(json.dumps(valid_spec))
    fmi_model = FmiModel(model=model)

    with pytest.raises(ValueError) as exc_info:
        _ = format_template_data(onnx=wind_generator_onnx, fmi_model=fmi_model, model_component=model)

    assert exc_info.match(
        re.escape(
            "The number of total output indexes for all outputs in the interface file (=202) \
cannot exceed the output size of the ml model (=130)"
        )
    )


def test_template_data_invalid_state_size(wind_generator_onnx: ONNXModel):
    valid_spec = {
        "name": "example",
        "version": "1.0",
        "inputs": [
            {"name": "inputs", "description": "My inputs", "agentInputIndexes": ["0:2"], "isArray": True, "length": 2}
        ],
        "outputs": [
            {"name": "outputs", "description": "My outputs", "agentInputIndexes": ["0:2"], "isArray": True, "length": 2}
        ],
        "states": [
            {"agentOutputIndexes": ["2:200"]},
        ],
    }

    _, model = validate_interface_spec(json.dumps(valid_spec))
    assert model is not None

    fmi_model = FmiModel(model=model)

    with pytest.raises(ValueError) as exc_info:
        _ = format_template_data(onnx=wind_generator_onnx, fmi_model=fmi_model, model_component=model)

    assert exc_info.match(
        re.escape(
            "The number of total output indexes for all states in the interface file (=198) \
cannot exceed either the state input size (=130)"
        )
    )



from os import path
from conan import ConanFile
from conan.errors import ConanInvalidConfiguration
from conan.tools.build import check_min_cppstd
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps, cmake_layout
from conan.tools.env import VirtualRunEnv
from conan.tools.scm import Git, Version
from conan.tools.files import copy, load, update_conandata

required_conan_version = ">=2.0.15"


class CppFmuConan(ConanFile):
    name = "cppfmu"
    license = "MPL-2.0"
    author = "Lars T. Kyllingstad"
    description = "C++ wrapper for FMI co-simulation, v1 and v2"
    topics = ("Co-simulation")
    url = "https://github.com/viproma/cppfmu"
    settings = "os", "compiler", "build_type", "arch"
    package_type = "static-library"
    options = {
        "fPIC": [True, False],
        "use_fmi_version": [1, 2]
    }
    default_options = {
        "fPIC": True,
        "use_fmi_version": 2
    }

    @property
    def _min_cppstd(self):
        return 11

    @property
    def _compilers_minimum_version(self):
        return {
            "apple-clang": "5.0",
            "clang": "3.3",
            "gcc": "4.9",
            "msvc": "170",
        }

    def export(self):
        copy(self, "version.txt", self.recipe_folder, self.export_folder)
        git = Git(self, self.recipe_folder)
        scm_url, scm_commit = git.get_url_and_commit()
        update_conandata(self, {"sources": {"commit": scm_commit, "url": scm_url}})

    def set_version(self):
        self.version = \
            load(self, path.join(self.recipe_folder, "version.txt")).strip()

    def config_options(self):
        if self.settings.os == "Windows":
            del self.options.fPIC

    def layout(self):
        cmake_layout(self)

    def requirements(self):
        if self.options.use_fmi_version == 1:
            self.requires("fmi1/1.0.1", transitive_headers=True)
        else:
            self.requires("fmi2/2.0.4", transitive_headers=True)

    def validate(self):
        if self.settings.compiler.cppstd:
            check_min_cppstd(self, self._min_cppstd)
        minimum_version = self._compilers_minimum_version.get(str(self.settings.compiler), False)
        if minimum_version and Version(self.settings.compiler.version) < minimum_version:
            raise ConanInvalidConfiguration(
                f"{self.ref} requires C++{self._min_cppstd}, which your compiler does not support."
            )

    def source(self):
        git = Git(self)
        sources = self.conan_data["sources"]
        git.clone(url=sources["url"], target=".")
        git.checkout(commit=sources["commit"])

    def generate(self):
        tc = CMakeToolchain(self)
        if self.options.use_fmi_version == 1:
            tc.variables["CPPFMU_FMI_1"] = True
        else:
            tc.variables["CPPFMU_FMI_1"] = False
        tc.generate()
        deps = CMakeDeps(self)
        deps.generate()

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()
        cmake.ctest("--output-on-failure")

    def package(self):
        copy(self, "LICENCE.txt", self.source_folder,
             path.join(self.package_folder, "licenses"))
        cmake = CMake(self)
        cmake.install()

    def package_info(self):
        self.cpp_info.libs = ["cppfmu"]
        self.cpp_info.srcdirs = ["src"]
        if self.options.use_fmi_version == 1:
            self.output.info("Define fmi1")
            self.cpp_info.defines = ["CPPFMU_USE_FMI_1_0=1"]
            self.cpp_info.requires = ["fmi1::cosim"]
        else:
            self.cpp_info.requires = ["fmi2::fmi2"]



from __future__ import annotations

import logging
import os
import tempfile
from enum import Enum
from pathlib import Path

from mlfmu.utils import builder
from mlfmu.utils.path import find_default_file

__ALL__ = ["run", "MlFmuProcess"]

logger = logging.getLogger(__name__)


class MlFmuCommand(Enum):
    """Enum class for the different commands in the mlfmu process."""

    BUILD = "build"
    GENERATE = "codegen"
    COMPILE = "compile"

    @staticmethod
    def from_string(command_string: str) -> MlFmuCommand | None:
        matches = [command for command in MlFmuCommand if command.value == command_string]
        return matches[0] if matches else None


# run for mlfmu
def run(
    command: MlFmuCommand,
    interface_file: str | None,
    model_file: str | None,
    fmu_path: str | None,
    source_folder: str | None,
) -> None:
    """Run the mlfmu process.

    Run the mlfmu process with the given command and optional parameters.

    Parameters
    ----------
    command: MlFmuCommand
        which command in the mlfmu process that should be run
    interface_file: Optional[str]
        the path to the file containing the FMU interface. Will be inferred if not provided.
    model_file: Optional[str]
        the path to the ml model file. Will be inferred if not provided.
    fmu_path: Optional[str]
        the path to where the built FMU should be saved. Will be inferred if not provided.
    source_folder: Optional[str]
        the path to where the FMU source code is located. Will be inferred if not provided.
    """
    process = MlFmuProcess(
        command=command,
        source_folder=Path(source_folder) if source_folder is not None else None,
        interface_file=Path(interface_file) if interface_file is not None else None,
        ml_model_file=Path(model_file) if model_file is not None else None,
        fmu_output_folder=Path(fmu_path) if fmu_path is not None else None,
    )
    process.run()

    return


class MlFmuBuilder:
    """Builder for creating FMUs (Functional Mock-up Units) from machine learning models.

    This class is core to executing the different commands in the mlfmu process.
    """

    def __init__(
        self,
        fmu_name: str | None = None,
        interface_file: Path | None = None,
        ml_model_file: Path | None = None,
        source_folder: Path | None = None,
        fmu_output_folder: Path | None = None,
        build_folder: Path | None = None,
        root_directory: Path | None = None,
    ) -> None:
        """Initialize a new instance of the MlFmuBuilder class.

        Parameters
        ----------
        fmu_name : str | None, optional
            The name of the FMU., by default None
        interface_file : Path | None, optional
            The path to the interface JSON file., by default None
        ml_model_file : Path | None, optional
            The path to the machine learning model file., by default None
        source_folder : Path | None, optional
            The folder containing the source code for the FMU., by default None
        fmu_output_folder : Path | None, optional
            The folder where the built FMU will be saved., by default None
        build_folder : Path | None, optional
            The folder where the FMU will be built., by default None
        root_directory : Path | None, optional
            The root directory for the builder., by default None
        """
        self.fmu_name: str | None = fmu_name
        self.interface_file: Path | None = interface_file
        self.ml_model_file: Path | None = ml_model_file
        self.source_folder: Path | None = source_folder
        self.fmu_output_folder: Path | None = fmu_output_folder
        self.build_folder: Path | None = build_folder
        self.root_directory: Path = root_directory or Path.cwd()
        #: The temporary folder used for building the FMU.
        self.temp_folder: tempfile.TemporaryDirectory[str] = tempfile.TemporaryDirectory(prefix="mlfmu_")
        self.temp_folder_path: Path = Path(self.temp_folder.name)
        logger.debug(f"Created temp folder: {self.temp_folder_path}")

    def build(self) -> None:
        """Build an FMU from the machine learning model file and interface file.

        Builds an FMU from ml_model_file and interface_file and saves it to fmu_output_folder.
        If the paths to the necessary files and directories are not given the function
        will try to find files and directories that match the ones needed.

        Raises
        ------
        FileNotFoundError
            if ml_model_file or interface_file do not exists or is not set and cannot be easily inferred.
        ---
        """
        # sourcery skip: class-extract-method
        logger.debug("MLFmuBuilder: Start build")
        # specify folders and filenames for building
        self.source_folder = self.source_folder or self.default_build_source_folder()

        self.ml_model_file = self.ml_model_file or self.default_model_file()
        if self.ml_model_file is None:
            raise FileNotFoundError(
                "No model file was provided and no obvious model file found in current working directory (os.getcwd())"
            )
        if not self.ml_model_file.exists():
            raise FileNotFoundError(f"The given model file (={self.ml_model_file}) does not exist.")

        self.interface_file = self.interface_file or self.default_interface_file()
        if self.interface_file is None:
            raise FileNotFoundError(
                "No interface json file was provided and no obvious interface file found in current working directory."
            )
        if not self.interface_file.exists():
            raise FileNotFoundError(f"The given interface json file (={self.interface_file}) does not exist.")

        self.build_folder = self.build_folder or self.default_build_folder()
        self.fmu_output_folder = self.fmu_output_folder or self.default_fmu_output_folder()

        # create fmu files
        try:
            fmi_model = builder.generate_fmu_files(self.source_folder, self.ml_model_file, self.interface_file)
        except Exception:
            logger.exception("Exception when running generate_fmu_files: %s")

        self.fmu_name = fmi_model.name
        builder.build_fmu(
            fmu_src_path=self.source_folder / self.fmu_name,
            fmu_build_path=self.build_folder,
            fmu_save_path=self.fmu_output_folder,
        )
        logger.debug("MLFmuBuilder: Done with build")

    def generate(self) -> None:
        """Generate C++ source code and model description from the machine learning model file and interface file.

        Generates FMU C++ source code and model description from ml_model_file and interface_file
        and saves it to source_folder.

        If the paths to the necessary files and directories are not given the function
        will try to find files and directories that match the ones needed.

        Raises
        ------
        FileNotFoundError
            if ml_model_file or interface_file do not exists or is not set and cannot be easily inferred.
        """
        logger.debug("MLFmuBuilder: Start generate")
        # specify folders and filenames for generating
        self.source_folder = self.source_folder or self.default_generate_source_folder()

        self.ml_model_file = self.ml_model_file or self.default_model_file()
        if self.ml_model_file is None:
            raise FileNotFoundError(
                "No model file was provided and no obvious model file found in current working directory (os.getcwd())"
            )
        if not self.ml_model_file.exists():
            raise FileNotFoundError(f"The given model file (={self.ml_model_file}) does not exist.")

        self.interface_file = self.interface_file or self.default_interface_file()
        if self.interface_file is None:
            raise FileNotFoundError(
                "No interface json file was provided and no obvious interface file found in current working directory."
            )
        if not self.interface_file.exists():
            raise FileNotFoundError(f"The given interface json file (={self.interface_file}) does not exist.")

        # create fmu files
        try:
            fmi_model = builder.generate_fmu_files(self.source_folder, self.ml_model_file, self.interface_file)
            self.fmu_name = fmi_model.name
        except Exception:
            logger.exception("Exception when running generate_fmu_files: %s")

        logger.debug("MLFmuBuilder: Done with generate")

    def compile(self) -> None:
        """Compile the FMU from the C++ source code and model description.

        Compiles the FMU from the FMU C++ source code and model description contained in source_folder
        and saves it to fmu_output_folder.

        If the paths to the necessary files and directories are not given the function
        will try to find files and directories that match the ones needed.

        Raises
        ------
        FileNotFoundError
            if source_folder or fmu_name is not set and cannot be easily inferred.
        """
        logger.debug("MLFmuBuilder: Start compile")
        self.build_folder = self.build_folder or self.default_build_folder()

        self.fmu_output_folder = self.fmu_output_folder or self.default_fmu_output_folder()

        if self.fmu_name is None or self.source_folder is None:
            source_child_folder = self.default_compile_source_folder()
            if source_child_folder is None:
                raise FileNotFoundError(
                    "No valid FMU source directory found anywhere inside the current working directory "
                    f"or any given source path (={self.source_folder})."
                )
            self.fmu_name = source_child_folder.stem
            self.source_folder = source_child_folder.parent

        try:
            builder.build_fmu(
                fmu_src_path=self.source_folder / self.fmu_name,
                fmu_build_path=self.build_folder,
                fmu_save_path=self.fmu_output_folder,
            )
            logger.debug("MLFmuBuilder: Done with build (via compile)")
        except Exception:
            logger.exception("Error while running build_fmu: %s")
        logger.debug("MLFmuBuilder: Done with compile")

    def default_interface_file(self) -> Path | None:
        """Return the path to a interface json file inside self.root_directory if it can be inferred.

        Returns
        -------
        Path | None
            the path to the interface json file, or None.
        """
        return find_default_file(self.root_directory, "json", "interface")

    def default_model_file(self) -> Path | None:
        """Return the path to a ml model file inside self.root_directory if it can be inferred.

        Returns
        -------
        Path | None
            the path to the ml model file, or None.
        """
        return find_default_file(self.root_directory, "onnx", "model")

    def default_build_folder(self) -> Path:
        """Return the path to a build folder inside the temp_folder. Creates the temp_folder if it is not set.

        Returns
        -------
        Path
            the path to the build folder.
        """
        return self.temp_folder_path / "build"

    def default_build_source_folder(self) -> Path:
        """Return the path to a src folder inside the temp_folder. Creates the temp_folder if it is not set.

        Returns
        -------
        Path
            the path to the src folder.
        """
        return self.temp_folder_path / "src"

    def default_generate_source_folder(self) -> Path:
        """Return the path to the default source folder for the generate process.

        Returns
        -------
        Path
            the path to the default source folder for the generate process.
        """
        return self.root_directory

    def default_compile_source_folder(self) -> Path | None:
        """Return the path to the default source folder for the compile process.

        Searches inside self.source_folder and self.root_directory for a folder that contains a folder structure
        and files that is required to be valid ml fmu source code.

        Returns
        -------
        Path
            the path to the default source folder for the compile process.
        """
        search_folders: list[Path] = []
        if self.source_folder is not None:
            search_folders.append(self.source_folder)
        search_folders.append(self.root_directory)
        source_folder: Path | None = None
        # If source folder is not provided, try to find one in current folder that is compatible with the tool
        # I.e a folder that contains everything needed for compilation
        for current_folder in search_folders:
            for sub_folder, _, _ in os.walk(current_folder):
                try:
                    possible_source_folder = Path(sub_folder)
                    # If a fmu name is given and the candidate folder name does not match. Skip it!
                    if self.fmu_name is not None and possible_source_folder.stem != self.fmu_name:
                        continue
                    builder.validate_fmu_source_files(possible_source_folder)
                    source_folder = possible_source_folder
                    # If a match was found stop searching
                    break
                except Exception:
                    logger.exception("Exception when validating source folder: %s")
                    # Any folder that does not contain the correct folder structure and files
                    # needed for compilation will raise an exception
                    continue
            # If a match was found stop searching
            if source_folder is not None:
                break
        return source_folder

    def default_fmu_output_folder(self) -> Path:
        """Return the path to the default fmu output folder.

        Returns
        -------
        Path
            the path to the default fmu output folder.
        """
        return self.root_directory


class MlFmuProcess:
    """
    Represents the ML FMU process.

    This class encapsulates the functionality to run the ML FMU process in a self-terminated loop.
    It provides methods to control the execution of the process and manage the number of runs.
    """

    command: MlFmuCommand  #: The command to be executed by the process.
    builder: MlFmuBuilder  #: The builder object responsible for building the FMU.

    def __init__(
        self,
        command: MlFmuCommand,
        source_folder: Path | None = None,
        ml_model_file: Path | None = None,
        interface_file: Path | None = None,
        fmu_output_folder: Path | None = None,
    ) -> None:
        """Instantiate the MlFmuProcess.

        Parameters
        ----------
        command : MlFmuCommand
            The command to be executed by the process.
        source_folder : Path | None, optional
            The path to the source folder., by default None
        ml_model_file : Path | None, optional
            The path to the ML model file., by default None
        interface_file : Path | None, optional
            The path to the interface file., by default None
        fmu_output_folder : Path | None, optional
            The path to the FMU output folder., by default None
        """
        self._run_number: int = 0
        self._max_number_of_runs: int = 1
        self.terminate: bool = False

        self.command = command

        fmu_name: str | None = None
        build_folder: Path | None = None

        self.builder = MlFmuBuilder(
            fmu_name=fmu_name,
            interface_file=interface_file,
            ml_model_file=ml_model_file,
            source_folder=source_folder,
            fmu_output_folder=fmu_output_folder,
            build_folder=build_folder,
        )

    def run(self) -> None:
        """
        Run the mlfmu process.

        Runs the mlfmu process in a self-terminated loop.
        """
        # Run mlfmu process until termination is flagged
        while not self.terminate:
            try:
                self._run_process()
            except Exception:
                logger.exception("Exception in run_process for MlFmuProcess: %s")
                self.terminate = True
            self.terminate = self._run_number >= self._max_number_of_runs

        return

    def _run_process(self) -> None:
        """Execute a single run of the mlfmu process."""
        self._run_number += 1

        logger.debug(f"MlFmuProcess: Start run {self._run_number}")

        if self.command == MlFmuCommand.BUILD:
            self.builder.build()
        elif self.command == MlFmuCommand.GENERATE:
            self.builder.generate()
        elif self.command == MlFmuCommand.COMPILE:
            self.builder.compile()

        logger.debug(f"MlFmuProcess: Successfully finished run {self._run_number}")

        return



from __future__ import annotations

import warnings
from dataclasses import dataclass
from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING, Annotated, Any

from pydantic import BaseModel, ConfigDict, StringConstraints, model_validator
from pydantic.fields import Field

from mlfmu.utils.signals import range_list_expanded
from mlfmu.utils.strings import to_camel

if TYPE_CHECKING:
    from uuid import UUID


class FmiVariableType(str, Enum):
    """Enum for variable type."""

    REAL = "real"
    INTEGER = "integer"
    STRING = "string"
    BOOLEAN = "boolean"


class FmiCausality(str, Enum):
    """Enum for variable causality."""

    PARAMETER = "parameter"
    INPUT = "input"
    OUTPUT = "output"


class FmiVariability(str, Enum):
    """Enum for signal variability."""

    CONSTANT = "constant"
    FIXED = "fixed"
    TUNABLE = "tunable"
    DISCRETE = "discrete"
    CONTINUOUS = "continuous"


class BaseModelConfig(BaseModel):
    """Enables the alias_generator for all cases."""

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
    )


class Variable(BaseModelConfig):
    """Pydantic model representing a variable in an FMU component."""

    #: Unique name for the port.
    name: str = Field(
        description="Unique name for the port.",
        examples=["windSpeed", "windDirection"],
    )
    #: Data type as defined by FMI standard.
    #: Defaults to Real.
    type: FmiVariableType = Field(
        default=FmiVariableType.REAL,
        description="Data type as defined by FMI standard, defaults to Real.",
        examples=[FmiVariableType.REAL, FmiVariableType.INTEGER],
    )
    #: Short FMU variable description.
    #: Optional.
    description: str | None = Field(
        default=None,
        description="Short FMU variable description.",
    )
    #: Signal variability as defined by FMI.
    #: Optional.
    variability: FmiVariability | None = Field(
        default=None,
        description="Signal variability as defined by FMI.",
    )
    #: Initial value of the signal at time step 1.
    #: Type should match the variable type.
    #: Defaults to 0.
    start_value: float | str | bool | int | None = Field(
        default=0,
        description="Initial value of the signal at time step 1. Type should match the variable type.",
    )
    #: When dealing with an array signal, it is essential to specify the LENGTH parameter.
    #: Arrays are indexed starting from 0, and FMU signals will be structured as
    #: SIGNAL_NAME[0], SIGNAL_NAME[1], and so forth.
    #: Defaults to False.
    is_array: bool = Field(
        default=False,
        description=(
            "When dealing with an array signal, it is essential to specify the LENGTH parameter. "
            "Arrays are indexed starting from 0, and FMU signals will be structured as "
            "SIGNAL_NAME[0], SIGNAL_NAME[1], and so forth. "
            "By default, this feature is set to False."
        ),
    )
    #: Defines the number of entries in the signal if the signal is array.
    #: Optional.
    length: int | None = Field(
        default=None,
        description="Defines the number of entries in the signal if the signal is array.",
        examples=[3, 5],
    )


class InternalState(BaseModelConfig):
    """Pydantic model representing an internal state of an FMU component."""

    #: Unique name for the state. Only needed if start_value is set (!= None).
    #: Initialization FMU parameters will be generated using this name.
    #: Optional.
    name: str | None = Field(
        default=None,
        description=(
            "Unique name for state. Only needed if start_value is set (!= None). "
            "Initialization FMU parameters will be generated using this name"
        ),
        examples=["initialWindSpeed", "initialWindDirection"],
    )
    #: Short description of the FMU variable.
    #: Optional.
    description: str | None = Field(
        default=None,
        description="Short FMU variable description.",
    )
    #: The default value of the parameter used for initialization.
    #: If this field is set, parameters for initialization will be automatically generated for these states.
    #: Optional.
    start_value: float | None = Field(
        default=None,
        description=(
            "The default value of the parameter used for initialization. "
            "If this field is set parameters for initialization will be automatically "
            "generated for these states."
        ),
    )
    #: The name of an input or parameter in the same model interface that should be used to initialize this state.
    #: Optional.
    initialization_variable: str | None = Field(
        default=None,
        description=(
            "The name of a an input or parameter in the same model interface "
            "that should be used to initialize this state."
        ),
    )
    #: Index or range of indices of agent (ONNX model) outputs that will be stored as internal states
    #: and will be fed as inputs in the next time step.
    #: Note: the FMU signal and the ONNX (agent) outputs need to have the same length.
    #: Defaults to an empty list.
    agent_output_indexes: list[
        Annotated[
            str,
            StringConstraints(strip_whitespace=True, to_upper=True, pattern=r"^(\d+|\d+:\d+)$"),
        ]
    ] = Field(
        default=[],
        description=(
            "Index or range of indices of agent outputs that will be stored as internal states "
            "and will be fed as inputs in the next time step. "
            "Note: the FMU signal and the agent outputs need to have the same length."
        ),
        examples=["10", "10:20", "30"],
    )

    # TODO @KristofferSkare: Change return type to `Self` (from `typing`module)
    #      once we drop support for Python 3.10
    #      (see https://docs.python.org/3/library/typing.html#typing.Self)
    #      CLAROS, 2024-10-15
    @model_validator(mode="after")
    def check_only_one_initialization(self) -> InternalState:
        """
        Check if only one state initialization method is used at a time.

        Raises a ValueError if multiple state initialization methods are used simultaneously.

        Returns
        -------
            self: The FMU component instance.

        Raises
        ------
            ValueError: If initialization_variable is set and either start_value or name is also set.
            ValueError: If name is set without start_value being set.
            ValueError: If start_value is set without name being set.
        """
        init_var = self.initialization_variable is not None
        name = self.name is not None
        start_value = self.start_value is not None

        if init_var and (start_value or name):
            raise ValueError(
                "Only one state initialization method is allowed to be used at a time: "
                "initialization_variable cannot be set if either start_value or name is set."
            )
        if (not start_value) and name:
            raise ValueError(
                "name is set without start_value being set. "
                "Both fields need to be set for the state initialization to be valid."
            )
        if start_value and (not name):
            raise ValueError(
                "start_value is set without name being set. "
                "Both fields need to be set for the state initialization to be valid."
            )
        return self


class InputVariable(Variable):
    """
    Represents an input variable for an FMU component.

    Examples
    --------
        An example of `agent_input_indexes` can be ["10", "10:20", "30"].
    """

    #: Index or range of indices of ONNX (agent) inputs to which this FMU signal shall be linked to.
    #: Note: The FMU signal and the ONNX (agent) inputs need to have the same length.
    #: Defaults to an empty list.
    agent_input_indexes: list[
        Annotated[
            str,
            StringConstraints(
                strip_whitespace=True,
                to_upper=True,
                pattern=r"^(\d+|\d+:\d+)$",
            ),
        ]
    ] = Field(
        default=[],
        description=(
            "Index or range of indices of agent inputs to which this FMU signal shall be linked to. "
            "Note: the FMU signal and the agent inputs need to have the same length."
        ),
        examples=["10", "10:20", "30"],
    )


class OutputVariable(Variable):
    """
    Represents an output variable in the FMU component.

    Examples
    --------
        An example of `agent_output_indexes` can be ["10", "10:20", "30"].
    """

    #: Index or range of indices of agent outputs that will be linked to this output signal.
    #: Note: The FMU signal and the agent outputs need to have the same length.
    #: Defaults to an empty list.
    agent_output_indexes: list[
        Annotated[
            str,
            StringConstraints(
                strip_whitespace=True,
                to_upper=True,
                pattern=r"^(\d+|\d+:\d+)$",
            ),
        ]
    ] = Field(
        default=[],
        description=(
            "Index or range of indices of agent outputs that will be linked to this output signal. "
            "Note: The FMU signal and the agent outputs need to have the same length."
        ),
        examples=["10", "10:20", "30"],
    )


# TODO @KristofferSkare: The `FmiInputVariable` and `FmiOutputVariable` classes are marked as `@dataclass`,
#      but they inherit from (at its very root) pydantic's `BaseModel`.
#      Let's discuss whether we can possibly remove the `@dataclass` decorator.
#      Then also the manual `__init__` method in `FmiInputVariable` and `FmiOutputVariable` could be removed.
#      CLAROS, 2024-10-15
@dataclass
class FmiInputVariable(InputVariable):
    """Data class representing an input variable in an FMI component."""

    causality: FmiCausality  #: The causality of the input variable.
    variable_references: list[int]  #: The list of variable references associated with the input variable.
    #: List of state initialization indexes for ONNX model - concerns mapping of FMU input variables to ONNX states.
    agent_state_init_indexes: list[list[int]] = []  # noqa: RUF008

    def __init__(self, **kwargs: Any) -> None:  # noqa: ANN401
        """Create an FMI input variable.

        Parameters
        ----------
        causality : FmiCausality, optional
            Causality of the input variable., by default FmiCausality.INPUT
        variable_references : list[int], optional
            List of variable references associated with the input variable., by default []
        """
        super().__init__(**kwargs)
        self.causality = kwargs.get("causality", FmiCausality.INPUT)
        self.variable_references = kwargs.get("variable_references", [])


@dataclass
class FmiOutputVariable(OutputVariable):
    """Data class representing an output variable in an FMI component."""

    causality: FmiCausality  #: The causality of the output variable.
    variable_references: list[int]  #: The list of variable references associated with the output variable.

    def __init__(self, **kwargs: Any) -> None:  # noqa: ANN401
        """Create an FMI output variable.

        Parameters
        ----------
        causality : FmiCausality, optional
            Causality of the output variable., by default FmiCausality.OUTPUT
        variable_references : list[int], optional
            List of variable references associated with the output variable., by default []
        """
        super().__init__(**kwargs)
        self.causality = kwargs.get("causality", FmiCausality.OUTPUT)
        self.variable_references = kwargs.get("variable_references", [])


@dataclass
class FmiVariable:
    """Data class representing a variable in an FMU component."""

    name: str = ""  #: The name of the variable.
    variable_reference: int = 0  #: The reference ID of the variable.
    type: FmiVariableType = FmiVariableType.REAL  #: The type of the variable.
    start_value: bool | str | int | float = 0  #: The initial value of the variable.
    causality: FmiCausality = FmiCausality.INPUT  #: The causality of the variable.
    description: str = ""  #: The description of the variable.
    variability: FmiVariability = FmiVariability.CONTINUOUS  #: The variability of the variable.


def _create_fmu_signal_example() -> Variable:
    """
    Create an example FMU signal variable.

    Returns
    -------
    Variable
        An instance of the Variable class representing the FMU signal variable.
    """
    return Variable(
        name="dis_yx",
        type=FmiVariableType.REAL,
        description=None,
        start_value=None,
        length=None,
        variability=None,
    )


class ModelComponent(BaseModelConfig):
    """Pydantic model representing a simulation model component.

    Used to generate the JSON schema for the model interface.
    Defines the structure of the FMU and how the inputs and outputs of the ONNX model
    correspond to the FMU variables.
    """

    #: The name of the simulation model.
    name: str = Field(
        description="The name of the simulation model.",
    )
    #: The version number of the model.
    #: Defaults to "0.0.1".
    version: str = Field(
        default="0.0.1",
        description="The version number of the model.",
    )
    #: Name or email of the model's author.
    #: Optional.
    author: str | None = Field(
        default=None,
        description="Name or email of the model's author.",
    )
    #: Short description of the model.
    #: Defaults to an empty string.
    description: str | None = Field(
        default="",
        description="Short description of the model.",
    )
    #: Copyright line for use in full license text.
    #: Optional.
    copyright: str | None = Field(
        default=None,
        description="Copyright line for use in full license text.",
    )
    #: License text or file name (relative to source files).
    #: Optional.
    license: str | None = Field(
        default=None,
        description="License text or file name (relative to source files)",
    )
    #: List of input signals of the simulation model.
    #: Defaults to an empty list.
    inputs: list[InputVariable] = Field(
        default=[],
        description="List of input signals of the simulation model.",
        examples=[[_create_fmu_signal_example()]],
    )
    #: List of output signals of the simulation model.
    #: Defaults to an empty list.
    outputs: list[OutputVariable] = Field(
        default=[],
        description="List of output signals of the simulation model.",
        examples=[[_create_fmu_signal_example()]],
    )
    #: List of parameter signals of the simulation model.
    #: Defaults to an empty list.
    parameters: list[InputVariable] = Field(
        default=[],
        description="List of parameter signals of the simulation model.",
        examples=[[_create_fmu_signal_example()]],
    )
    #: Internal states that will be stored in the simulation model's memory,
    #: these will be passed as inputs to the agent in the next time step.
    #: Defaults to an empty list.
    states: list[InternalState] = Field(
        default=[],
        description=(
            "Internal states that will be stored in the simulation model's memory, "
            "these will be passed as inputs to the agent in the next time step."
        ),
    )
    #: Whether the agent consumes time data from co-simulation algorithm.
    #: Defaults to False.
    uses_time: bool | None = Field(
        default=False,
        description="Whether the agent consumes time data from co-simulation algorithm.",
    )
    #: Whether variables are allowed to be reused for state initialization when
    #: initialization_variable is used for state initialization.
    #: If set to true the variable referred to in initialization_variable will be repeated
    #: for the state initialization until the entire state is initialized.
    #: Defaults to False.
    state_initialization_reuse: bool = Field(
        default=False,
        description=(
            "Whether variables are allowed to be reused for state initialization "
            "when initialization_variable is used for state initialization. "
            "If set to true the variable referred to in initialization_variable will be repeated "
            "for the state initialization until the entire state is initialized."
        ),
    )


class FmiModel:
    """Represents an FMU model with its associated properties and variables."""

    def __init__(self, model: ModelComponent) -> None:
        """Initialize the FmiModel object with a ModelComponent object.

        Parameters
        ----------
        model : ModelComponent
            FMU component compliant with FMISlave
        """
        # Assign model specification to a valid FMU component compliant with FMISlave
        self.name: str = model.name  #: The name of the FMU model.
        self.guid: UUID | None = None  #: The globally unique identifier of the FMU model.
        self.inputs: list[FmiInputVariable] = []  #: The list of input variables for the FMU model.
        self.outputs: list[FmiOutputVariable] = []  #: The list of output variables for the FMU model.
        self.parameters: list[FmiInputVariable] = []  #: The list of parameter variables for the FMU model.
        self.author: str | None = model.author  #: The author of the FMU model.
        self.version: str = model.version or "0.0.1"  # The version of the FMU model.
        self.description: str | None = model.description  #: The description of the FMU model.
        self.copyright: str | None = model.copyright  #: The copyright information of the FMU model.
        self.license: str | None = model.license  #: The license information of the FMU model.
        #: Whether the FMU model reuses state initialization.
        self.state_initialization_reuse: bool = model.state_initialization_reuse

        self.add_variable_references(model.inputs, model.parameters, model.outputs)
        self.add_state_initialization_parameters(model.states)

    def add_variable_references(
        self,
        inputs: list[InputVariable],
        parameters: list[InputVariable],
        outputs: list[OutputVariable],
    ) -> None:
        """Assign variable references to inputs, parameters and outputs from user interface to the FMU model class.

        Parameters
        ----------
        inputs : list[InputVariable]
            List of input variables from JSON interface.
        parameters : list[InputVariable]
            List of model parameters from JSON interface.
        outputs : list[OutputVariable]
            List of output variables from JSON interface.
        """
        current_var_ref = 0
        fmu_inputs: list[FmiInputVariable] = []
        fmu_parameters: list[FmiInputVariable] = []
        fmu_outputs: list[FmiOutputVariable] = []
        fmi_variable: FmiInputVariable | FmiOutputVariable

        var: InputVariable | OutputVariable

        for var in inputs:
            var_port_refs = []

            if var.is_array:
                # If array then allocate space for every element
                vector_port_length = var.length or 1
                var_port_refs = list(range(current_var_ref, current_var_ref + vector_port_length))
            else:
                var_port_refs = [current_var_ref]

            # Set current variable reference based on number of ports used by this input (array or scalar port)
            current_var_ref = current_var_ref + len(var_port_refs)
            fmi_variable = FmiInputVariable(
                causality=FmiCausality.INPUT,
                variable_references=var_port_refs,
                **var.__dict__,
            )
            fmu_inputs.append(fmi_variable)

        for var in parameters:
            var_port_refs = []

            if var.is_array:
                # If array then allocate space for every element
                vector_port_length = var.length or 1
                var_port_refs = list(range(current_var_ref, current_var_ref + vector_port_length))
            else:
                var_port_refs = [current_var_ref]

            # Set current variable reference based on number of ports used by this input (array or scalar port)
            current_var_ref = current_var_ref + len(var_port_refs)
            fmi_variable = FmiInputVariable(
                causality=FmiCausality.PARAMETER,
                variable_references=var_port_refs,
                **var.__dict__,
            )
            fmu_parameters.append(fmi_variable)

        for var in outputs:
            var_port_refs = []

            if var.is_array:
                # If array then allocate space for every element
                vector_port_length = var.length or 1
                var_port_refs = list(range(current_var_ref, current_var_ref + vector_port_length))
            else:
                var_port_refs = [current_var_ref]

            # Set current variable reference based on number of ports used by this input (array or scalar port)
            current_var_ref = current_var_ref + len(var_port_refs)
            fmi_variable = FmiOutputVariable(
                causality=FmiCausality.OUTPUT,
                variable_references=var_port_refs,
                **var.__dict__,
            )
            fmu_outputs.append(fmi_variable)

        self.inputs = fmu_inputs
        self.outputs = fmu_outputs
        self.parameters = fmu_parameters

    def add_state_initialization_parameters(
        self,
        states: list[InternalState],
    ) -> None:
        """Generate or modify FmuInputVariables for initialization of states.

        Generates or modifies FmuInputVariables for initialization of states for the InternalState objects
        that have set start_value and name or have set initialization_variable.
        Any generated parameters are appended to self.parameters.

        Parameters
        ----------
        states : list[InternalState]
            List of states from JSON interface.

        Raises
        ------
        ValueError
            If variables with same name are found. Variables must have a unique name.
        ValueError
            If no FMU variables were found for use for initialization.
        ValueError
            If a state has a state_value != None without having a name.
        """
        init_parameters: list[FmiInputVariable] = []

        # TODO @KristofferSkare: Biggest used value reference + 1, will this always be correct?
        value_reference_start = self.get_total_variable_number()
        current_state_index_state = 0
        for i, state in enumerate(states):
            length = len(range_list_expanded(state.agent_output_indexes))
            if state.initialization_variable is not None:
                variable_name = state.initialization_variable
                variable_name_input_index = [i for i, inp in enumerate(self.inputs) if inp.name == variable_name]
                variable_name_parameter_index = [
                    i for i, param in enumerate(self.parameters) if param.name == variable_name
                ]
                if len(variable_name_input_index) + len(variable_name_parameter_index) > 1:
                    raise ValueError(
                        f"Found {len(variable_name_input_index) + len(variable_name_parameter_index)} "
                        f"FMU inputs or parameters with same name (={variable_name}) "
                        "when trying to use for state initialization. Variables must have a unique name."
                    )

                if len(variable_name_input_index) + len(variable_name_parameter_index) == 0:
                    raise ValueError(
                        "Did not find any FMU variables for use for initialization "
                        f"with name={variable_name} for state with agent_output_indexes={state.agent_output_indexes}."
                    )
                agent_state_init_indexes = list(range(current_state_index_state, current_state_index_state + length))

                if len(variable_name_input_index) == 1:
                    self.inputs[variable_name_input_index[0]].agent_state_init_indexes.append(agent_state_init_indexes)
                if len(variable_name_parameter_index) == 1:
                    self.parameters[variable_name_parameter_index[0]].agent_state_init_indexes.append(
                        agent_state_init_indexes
                    )

            elif state.start_value is not None:
                if state.name is None:
                    raise ValueError(
                        f"State with index {i} has state_value (!= None) without having a name. "
                        "Either give it a name or set start_value = None"
                    )
                value_references = list(range(value_reference_start, value_reference_start + length))
                is_array = length > 1
                init_param = FmiInputVariable(
                    name=state.name,
                    description=state.description,
                    start_value=state.start_value,
                    variability=FmiVariability.FIXED,
                    type=FmiVariableType.REAL,
                    causality=FmiCausality.PARAMETER,
                    variable_references=value_references,
                    length=length,
                    is_array=is_array,
                    agent_input_indexes=[],
                    agent_state_init_indexes=[
                        list(range(current_state_index_state, current_state_index_state + length))
                    ],
                )
                init_parameters.append(init_param)
                value_reference_start += length
            current_state_index_state += length
        self.parameters = [*self.parameters, *init_parameters]

    def format_fmi_variable(
        self,
        var: FmiInputVariable | FmiOutputVariable,
    ) -> list[FmiVariable]:
        """Get an inclusive list of variables from an interface variable definition.

        Vectors are separated as N number of signals, being N the size of the array.

        Parameters
        ----------
        var : FmiInputVariable | FmiOutputVariable
            Interface variable definition with the variable references.

        Returns
        -------
        list[FmiVariable]
            A list of FMI formatted variables.
        """
        variables: list[FmiVariable] = []

        if var.is_array:
            for idx, var_ref in enumerate(var.variable_references):
                # Create port names that contain the index starting from 1. E.i signal[0], signal[1] ...
                name = f"{var.name}[{idx}]"
                fmi_var = FmiVariable(
                    name=name,
                    variable_reference=var_ref,
                    causality=var.causality,
                    description=var.description or "",
                    variability=var.variability
                    or (
                        FmiVariability.CONTINUOUS if var.causality != FmiCausality.PARAMETER else FmiVariability.TUNABLE
                    ),
                )
                variables.append(fmi_var)
        else:
            # Create a single variable in case it's not a vector port
            fmi_var = FmiVariable(
                name=var.name,
                variable_reference=var.variable_references[0],
                causality=var.causality,
                description=var.description or "",
                variability=var.variability
                or (FmiVariability.CONTINUOUS if var.causality != FmiCausality.PARAMETER else FmiVariability.TUNABLE),
                start_value=var.start_value if var.start_value is not None else 0,
                type=var.type or FmiVariableType.REAL,
            )
            variables.append(fmi_var)

        return variables

    def get_fmi_model_variables(self) -> list[FmiVariable]:
        """Get a full list of all variables in the FMU, including each index of vector ports.

        Returns
        -------
        list[FmiVariable]
            List of all variables in the FMU.
        """
        variables: list[FmiInputVariable | FmiOutputVariable]
        variables = [*self.inputs, *self.parameters, *self.outputs]
        fmi_variables = [self.format_fmi_variable(var) for var in variables]

        flat_vars = [var_j for var_i in fmi_variables for var_j in var_i]
        return flat_vars

    def get_template_mapping(
        self,
    ) -> tuple[list[tuple[int, int]], list[tuple[int, int]], list[tuple[int, int]]]:
        """Calculate the index to value reference mapping between onnx inputs/outputs/state to fmu variables.

        Returns
        -------
        tuple[list[tuple[int, int]], list[tuple[int, int]], list[tuple[int, int]]]
            Tuple of lists of mappings between onnx indexes to fmu variables.
            (input_mapping, output_mapping, state_init_mapping)
        """
        # Input and output mapping in the form of agent index and fmu variable reference pairs
        input_mapping: list[tuple[int, int]] = []
        output_mapping: list[tuple[int, int]] = []
        state_init_mapping: list[tuple[int, int]] = []

        for inp in self.inputs + self.parameters:
            input_indexes = range_list_expanded(inp.agent_input_indexes)
            input_mapping.extend(
                (input_index, inp.variable_references[variable_index])
                for variable_index, input_index in enumerate(input_indexes)
            )
            num_variable_references = len(inp.variable_references)
            for state_init_indexes in inp.agent_state_init_indexes:
                num_state_init_indexes = len(state_init_indexes)
                for variable_index, state_init_index in enumerate(state_init_indexes):
                    _variable_index = variable_index
                    if _variable_index >= num_variable_references:
                        if not self.state_initialization_reuse:
                            warnings.warn(
                                f"Too few variables in {inp.name} (={num_variable_references}) "
                                f"to initialize all states (={num_state_init_indexes}). "
                                "To initialize all states set `state_initialization_reuse=true` in interface json "
                                f"or provide a variable with length >={num_state_init_indexes}",
                                stacklevel=1,
                            )
                            break
                        _variable_index = _variable_index % num_variable_references
                    state_init_mapping.append((state_init_index, inp.variable_references[_variable_index]))

        for out in self.outputs:
            output_indexes = range_list_expanded(out.agent_output_indexes)
            output_mapping.extend(
                (output_index, out.variable_references[variable_index])
                for variable_index, output_index in enumerate(output_indexes)
            )
        input_mapping = sorted(input_mapping, key=lambda inp: inp[0])
        output_mapping = sorted(output_mapping, key=lambda out: out[0])
        return input_mapping, output_mapping, state_init_mapping

    def get_total_variable_number(self) -> int:
        """Calculate the total number of variables including every index of vector ports.

        Returns
        -------
        int
            The total number of variables.
        """
        all_fmi_variables: list[FmiInputVariable | FmiOutputVariable] = [
            *self.inputs,
            *self.parameters,
            *self.outputs,
        ]
        num_variables = reduce(
            lambda prev, current: prev + len(current.variable_references),
            all_fmi_variables,
            0,
        )
        return num_variables



def range_list_expanded(list_of_ranges: list[str]) -> list[int]:
    """
    Expand ranges specified in interface. They should be formatted as [starts_index:end_index].

    Args
    ----
        list_of_ranges (List[str]): List of indexes or ranges.

    Returns
    -------
        A list of all indexes covered in ranges or individual indexes
    """
    indexes: list[int] = []

    for val in list_of_ranges:
        if ":" in val:
            indexes.extend(range(*[int(num) for num in val.split(":")]))
        else:
            indexes.append(int(val))

    return indexes



#!/usr/bin/env python

import argparse
import logging
import textwrap
from pathlib import Path

from mlfmu.api import MlFmuCommand, run
from mlfmu.utils.logging import configure_logging

logger = logging.getLogger(__name__)


def _argparser() -> argparse.ArgumentParser:
    """
    Create and return an ArgumentParser object for parsing command line arguments.

    Returns
    -------
        argparse.ArgumentParser: The ArgumentParser object.
    """
    parser = argparse.ArgumentParser(
        prog="mlfmu",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""\
                                    mlfmu is a command line tool to build FMUs from ONNX ML models.
                                    Check the README and docs for more info.
                                    You can also run `mlfmu <command> --help` for more info on a specific command."""),
        epilog=textwrap.dedent("""\
                               This tool utilizes cppfmu, source code is available at: https://github.com/viproma/cppfmu
                               _________________mlfmu___________________
                               """),
        prefix_chars="-",
        add_help=True,
    )

    common_args_parser = argparse.ArgumentParser(add_help=False)

    console_verbosity = common_args_parser.add_mutually_exclusive_group(required=False)

    _ = console_verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=("console output will be quiet."),
        default=False,
    )

    _ = console_verbosity.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help=("console output will be verbose."),
        default=False,
    )

    _ = common_args_parser.add_argument(
        "-l",
        "--log",
        action="store",
        type=str,
        help=(
            "name of log file. If specified, this will activate logging to file. "
            "If not, it does not log to file, only console."
        ),
        default=None,
        required=False,
    )

    _ = common_args_parser.add_argument(
        "-ll",
        "--log-level",
        action="store",
        type=str,
        help="log level applied to logging to file. Default: WARNING.",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        required=False,
    )

    # Create a sub parser for each command
    sub_parsers = parser.add_subparsers(
        dest="command",
        title="Available commands",
        metavar="command",
        required=True,
    )

    # Main command
    # build command to go from config to compiled fmu
    build_parser = sub_parsers.add_parser(
        MlFmuCommand.BUILD.value,
        help="Build FMU from interface and model files",
        parents=[common_args_parser],
        add_help=True,
    )

    # Add options for build command
    _ = build_parser.add_argument(
        "-i",
        "--interface-file",
        type=str,
        help="JSON file describing the FMU following schema",
    )
    _ = build_parser.add_argument(
        "-m",
        "--model-file",
        type=str,
        help="ONNX file containing the ML Model",
    )
    _ = build_parser.add_argument(
        "-f",
        "--fmu-path",
        type=str,
        help="Path to where the built FMU should be saved",
    )

    # Split the main build command into steps for customization
    # generate-code command to go from config to generated fmu source code
    code_generation_parser = sub_parsers.add_parser(
        MlFmuCommand.GENERATE.value,
        help="Generate FMU source code from interface and model files",
        parents=[common_args_parser],
        add_help=True,
    )

    # Add options for code generation command
    _ = code_generation_parser.add_argument(
        "--interface-file",
        type=str,
        help="json file describing the FMU following schema (e.g. interface.json).",
    )
    _ = code_generation_parser.add_argument(
        "--model-file",
        type=str,
        help="onnx file containing the ML Model (e.g. example.onnx).",
    )
    _ = code_generation_parser.add_argument(
        "--fmu-source-path",
        help=(
            "Path to where the generated FMU source code should be saved. "
            "Given path/to/folder the files can be found in path/to/folder/[FmuName]"
        ),
    )

    # build-code command to go from fmu source code to compiled fmu
    build_code_parser = sub_parsers.add_parser(
        MlFmuCommand.COMPILE.value,
        help="Build FMU from FMU source code",
        parents=[common_args_parser],
        add_help=True,
    )

    # Add option for fmu compilation
    _ = build_code_parser.add_argument(
        "--fmu-source-path",
        type=str,
        help=(
            "Path to the folder where the FMU source code is located. "
            "The folder needs to have the same name as the FMU. E.g. path/to/folder/[FmuName]"
        ),
    )
    _ = build_code_parser.add_argument(
        "--fmu-path",
        type=str,
        help="Path to where the built FMU should be saved.",
    )

    return parser


def main() -> None:
    """
    Entry point for console script as configured in setup.cfg.

    Runs the command line interface and parses arguments and options entered on the console.
    """
    parser = _argparser()
    args = parser.parse_args()

    # Configure Logging
    # ..to console
    log_level_console: str = "WARNING"
    if any([args.quiet, args.verbose]):
        log_level_console = "ERROR" if args.quiet else log_level_console
        log_level_console = "INFO" if args.verbose else log_level_console
    # ..to file
    log_file: Path | None = Path(args.log) if args.log else None
    log_level_file: str = args.log_level
    configure_logging(log_level_console, log_file, log_level_file)
    logger.info("Logging to file: %s", log_file)

    command: MlFmuCommand | None = MlFmuCommand.from_string(args.command)

    if command is None:
        raise ValueError(
            f"The given command (={args.command}) does not match any of the existing commands "
            f"(={[command.value for command in MlFmuCommand]})."
        )

    interface_file = args.interface_file if "interface_file" in args else None
    model_file = args.model_file if "model_file" in args else None
    fmu_path = args.fmu_path if "fmu_path" in args else None
    source_folder = args.fmu_source_path if "fmu_source_path" in args else None

    # Invoke API
    try:
        run(
            command=command,
            interface_file=interface_file,
            model_file=model_file,
            fmu_path=fmu_path,
            source_folder=source_folder,
        )
    except Exception:
        logger.exception("Unhandled exception in run: %s")


if __name__ == "__main__":
    try:
        main()
    except Exception:
        logger.exception("Unhandled exception in main: %s")



import datetime
import logging
from importlib import metadata
from xml.etree.ElementTree import Element, ElementTree, SubElement, indent

from mlfmu.types.fmu_component import (
    FmiCausality,
    FmiModel,
    FmiVariability,
    FmiVariable,
)

logger = logging.getLogger(__name__)


def requires_start(var: FmiVariable) -> bool:
    """
    Test if a variable requires a start attribute.

    Returns
    -------
        True if successful, False otherwise
    """
    return var.causality in (FmiCausality.INPUT, FmiCausality.PARAMETER) or var.variability == FmiVariability.CONSTANT


def generate_model_description(fmu_model: FmiModel) -> ElementTree:
    """
    Generate FMU modelDescription as XML.

    Args
    ----
        fmu_model (FmiModel): Object representation of FMI slave instance

    Returns
    -------
        xml.etree.TreeElement.Element: modelDescription XML representation.
    """
    t = datetime.datetime.now(datetime.timezone.utc)
    date_str = t.isoformat(timespec="seconds")
    TOOL_VERSION = metadata.version("mlfmu")  # noqa: N806

    # Root <fmiModelDescription> tag
    model_description = {
        "fmiVersion": "2.0",
        "modelName": fmu_model.name,
        "guid": f"{fmu_model.guid!s}" if fmu_model.guid is not None else "@FMU_UUID@",
        "version": fmu_model.version,
        "generationDateAndTime": date_str,
        "variableNamingConvention": "structured",
        "generationTool": f"MLFMU {TOOL_VERSION}",
    }

    # Optional props
    if fmu_model.copyright is not None:
        model_description["copyright"] = fmu_model.copyright
    if fmu_model.license is not None:
        model_description["license"] = fmu_model.license
    if fmu_model.author is not None:
        model_description["author"] = fmu_model.author
    if fmu_model.description is not None:
        model_description["description"] = fmu_model.description

    root = Element("fmiModelDescription", model_description)

    # <CoSimulation> tag options
    cosim_options = {
        "modelIdentifier": fmu_model.name,
        "canHandleVariableCommunicationStepSize": "true",
    }
    _ = SubElement(root, "CoSimulation", attrib=cosim_options)

    # <ModelVariables> tag -> Append inputs/parameters/outputs
    variables = SubElement(root, "ModelVariables")

    # <ModelStructure> tag with <Outputs> tab inside --> Append all outputs
    model_structure = SubElement(root, "ModelStructure")
    outputs = SubElement(model_structure, "Outputs")

    for var in fmu_model.get_fmi_model_variables():
        # XML variable attributes
        var_attrs = {
            "name": var.name,
            "valueReference": str(var.variable_reference),
            "causality": var.causality.value,
            "description": var.description or "",
            "variability": var.variability.value if var.variability else FmiVariability.CONTINUOUS.value,
        }
        var_elem = SubElement(variables, "ScalarVariable", var_attrs)

        var_type_attrs = {}
        if requires_start(var):
            var_type_attrs["start"] = str(var.start_value)

        # FMI variable type element
        _ = SubElement(var_elem, var.type.value.capitalize(), var_type_attrs)

        # Appending output to <Outputs> inside <ModelStructure>
        if var.causality == FmiCausality.OUTPUT:
            _ = SubElement(outputs, "Unknown", {"index": str(var.variable_reference)})

    # Create XML tree containing root element and pretty format its contents
    xml_tree = ElementTree(root)
    indent(xml_tree, space="\t", level=0)
    return xml_tree



import logging

from mlfmu.utils.interface import publish_interface_schema

logger = logging.getLogger(__name__)


def main() -> None:
    """Start the publishing process for the mlfmu interface docs, by calling the publish_interface_schema function."""
    logger.info("Start publish-interface-docs.py")
    publish_interface_schema()
